#ifndef P1CAMERA_CAMERA_HPP
#define P1CAMERA_CAMERA_HPP

#include <string>
#include <vector>
#include "P1CameraTypes.hpp"
#include "P1CameraImageMeta.hpp"
#include "P1CameraImageChunk.hpp"
#include "P1CameraLiveViewConfiguration.hpp"

#include <C_P1CameraCommonLinkage.h>
#include <C_P1CameraCommonStructs.h>
#include <C_P1CameraCommonDeprecated.h>

namespace P1
{
namespace CameraSdk
{
    class CameraPimpl;
    class Notifications;

    /**
     * @brief Representation of a physical Phase One camera
     *
     * All interaction with camera are through objects of this class. You work with cameras using
     * two different schemes.
     *
     * #### Properties
     *
     * All Phase One camera expose a dynamic set of properties. These include *Aperture*, *Shutter Time*,
     * *ISO* etc. But properties can also represent other camera settings like *shutter mode* or the
     * dimension of the LiveView images.
     *
     * Common for all properties is that they define metadata like names and valid ranges.
     *
     * A camera exposes different properties at different times. This means that setting one property,
     * might cause other properties to disappear or become visible. Because of this dynamic nature, we do not
     * include an `enum` type with all @ref PropertyId. You must query for the name and purpose
     * of the properties to wish to use. And know that their existence is not guaranteed.
     *
     * #### Capturing & LiveView
     *
     * This class defines a set of methods to trigger image capturing on the camera, and streaming LiveView
     * images to SDK host computer. Also, you can control whether captured images should be transferred to the computer.
     *
     * Both captured and LiveView images are transferred in *data chunks*. The captured full resolution RAW images are large,
     * and transferring them in chunks saves memory. You can control the size of each chunk, so you could allocate a giant chunk
     * to transfer the entire image in one go.
     *
     * The format of the captured images are IIQ files, that can be written directly to disk. However, you must account for
     * data chunks arrive **un-ordered**. This means the received series of chunks and not a stream of bytes.
     * The end of the IIQ file might arrive in a chunk, before the begining. Each chunk will define its global byte offset.
     *
     * #### Creating & disposing Camera objects
     *
     * There is no (real) public `constructor` on this class. Since this class represents a association with physical cameras,
     * you cannot construct object on your own. Use the static method @ref OpenUsbCamera() to create cameras.
     *
     * Upon destructing an instance of this class, the connection to the physical camera is terminated and all associated
     * resources are freed.
     */
    class EXPORT_API Camera
    {
        friend class CameraPimpl;
        friend class Notifications;
    private:
        CameraPimpl* mImpl;

        /**
        * @brief Copy constructor deleted
        *
        * You should not duplicate the instance, since it represents a physical resource.
        * Use references or pointers (if you must) to pass camera objects around in your code.
        *
        * **The move contructor is available.**
        */
        Camera(const Camera& other) = delete;

        /**
        * @brief Assignment constructor deleted
        *
        * You should not duplicate the instance, since it represents a physical resource.
        * Use references or pointers (if you must) to pass camera objects around in your code.
        *
        * **The move assignment is available.**
        */
        Camera& operator=(const Camera& other) = delete;

        // Throws an SdkException if the camera is not open/valid. To be used for all methods requiring an open camera.
        void ThrowExceptionIfNotValid() const;

    public:

        /**
         * @brief Opens an USB Camera with the given serial-number. (Or any USB camera if the serial-number is omitted or empty).
         *
         * @param serial Optional: The serial of the camera to open. An empty string or `nullptr`, to open any camera.
         * @return A camera object, to control the requirested physical camera
         */
        static Camera OpenUsbCamera(const std::string& serial = std::string(), const std::string& password = std::string());


        /**
         * @brief Opens a TCP/IP camera with a given IP address
         *
         * @param ipAddress The IPv4 address of the camera
         * @param password An optional password, if camera is password protected
         * @param timeout An optional connection timeout in seconds. (0 is system's sockets default timeout)
         * @return A camera object, to control the requirested physical camera
         */
        static Camera OpenIpCamera (const std::string& ipAddress, const std::string& password = std::string(), const int timeoutSecs = 0);

        /**
         * @brief Returns a list of all USB camera *currently* attached to the system.
         *
         * Cameras come and go. The availble cameras might disappear right after this
         * method has returned. Therefore, opening a camera found using this method
         * might still fail.
         *
         * Cameras connected by TCP/IP _can not_ be discovered by this method!
         *
         * @return A list of cameras attached to your host system
         */
        static std::vector<CameraListElement> GetAvailableCameras();

        /**
        * @brief Create an empty dummy camera object - with no tether to real world camera
        *
        * This default contructor is meant allow initialization of an empty camera variable.
        * Calling any method on object created with this constructor will throw an exception!
        */
        Camera();

        /**
         * @brief Close camera connection and associated resources
         */
        ~Camera();

        /**
         * @brief Move camera from one context to another
         *
         * This is the *move constructor*, allowing move semantics on the camera objects.
         */
        Camera(Camera&& other);

        /**
         * @brief Assign a camera object to another variable
         *
         * This is the *move assignment*, allowing move semantics on the camera objects.
         */
        Camera& operator=(Camera&& other);

        /**
         * @brief Frees the camera handle, letting other processes or resources take over use of the camera.
         *
         * Closing the communication with the camera and causes any intialized states to shut down.
         * Likewise all event notifications from this camera are removed.
         */
        void Close();

        // Returns whether this camera is open.
        bool IsOpen() const;

        /**
         * @brief Get the unique id for this camera
         *
         * The Id is unique only for this runtime session.
         */
        int GetId();

        /**
         * @brief Wakes up any threads block in @ref WaitForImage or @ref WaitForLiveView calls
         *
         * To gracefullt terminate a multi-threaded application, use this method to force wake-up
         * threads there are waiting inside WaitFor..() calls.
         *
         * This method will force the calls to release the treads and return with a time-out exception.
         */
        void WakeUpWaitingImageThreads();


        /**
         * @brief Take a picture with the camera
         *
         * This method triggers the camera to take a picture, with its current settings and mode.
         *
         * If you with to receive the picture taken, you must first call @ref EnableImageReceiving
         * and set it to `true`. If not, the triggering will fail if there is no storage media in the camera.
         *
         * Be aware that actual triggering might also fail for a couple of other reasons. You need to listen for
         * the trigger event notification, to be sure that triggering occurred.
         */
        void TriggerCapture();

        /**
         * @brief Enable image captures are transmitted to the SDK
         *
         * Setting this to `true` tells the camera that the host computer is ready to receive captures.
         * The camera may choose to ignore this (and not transmit any captures), as the storage
         * configuration on the camera decides.
         *
         * @param enable `true` is enabled. `false` otherwise.
         */
        void EnableImageReceiving(bool enable);

        /**
         * @brief Registers this host's available storage capacity (in megabytes).
         *
         * If the host capacity falls below a certain threshold (depending on capture-settings),
         * the camera will reject new captures (depending on storage-settings).
         *
         * Initially, this setting is set to 1GB.
         */
        void SetHostStorageCapacity(uint32_t capacityMB);

        /**
         * @brief Wait for the next image to be captured / delivered from camera
         *
         * Before any images can be received with this method, you _must_ call EnableImageReceiving
         * to tell the camera to transfer images to you.
         *
         * The received image structure contains a std::shared_ptr that holds the raw data pointer.
         * This data is temporary, so you must copy it to a new location, either in memory or
         * on the file system. Do _not_ hold on to this pointer, since the memory it holds need to
         * be given back to the SDK subsystem.
         *
         * @param timeoutMs Defines an optional timeout (ms) - 0 is wait forever
         * @return const IIQImageFile A reference to an in-memory IIQ image file.
         */
        const IIQImageFile WaitForImage(uint32_t timeoutMs = 0);

        /**
         * @brief Turn capturing and transferring of LiveView images on
         *
         * To receive a stream of LiveView images, set this to `true`. This causes the camera to begin
         * feeding a stream of bitmap images to you host system.
         *
         * By default LiveView images are 1024 x 768 pixels, and 24 bit RGB. You can change these
         * settings by getting the LiveView configuration object @ref GetLiveViewConfiguration
         */
        void EnableLiveView();

        /**
         * @brief Turn capturing and transferring of LiveView images off
         */
        void DisableLiveView();

        /**
         * @brief Get (wait for) the lastest received live image
         *
         * Block and wait for the next received LiveView image from the camera. If
         * an LiveView image has already been received, it is returned immediately.
         *
         * The returned struct contains a pointer to the image's bitmap data. This memory is
         * the SDK's internal memory, and you should _not_ hold on to it. Copy the data either
         * to an in-memory location or to a file.
         *
         * Unlike captured images, LiveView frames does not arrive in a queue. Only the latest
         * live view image is preserved.
         *
         * @param timeoutMs The maximum amount of milliseconds to wait for a LiveView image to arrive
         * @return const LiveViewImage Struct that hold image data and relevant meta data
         */
        const LiveViewImage WaitForLiveView(uint32_t timeoutMs = 0);

        /**
         * @brief Get the LiveView configuration for the camera
         *
         * The returned object represents the LiveView configuration parameters on this camera. Changing
         * the parameters exposed by the object, will change the corrosponding LiveView parameters on
         * the camera.
         */
        LiveViewConfiguration& GetLiveViewConfiguration();

        // PROPERTIES

        /**
         * @brief Gets the current value of a given property
         *
         * The returned @ref PropertyValue object is a wrapper that encapsulate the property's type.
         * Since properties each can have different types, that is only known at runtime - all
         * property values are wrapped in the `PropertyValue` class.
         *
         * @param property The Id of the property to get
         * @return The value object of the property
         */
        PropertyValue GetProperty(const uint32_t property) const;

        /**
         * @brief Set a given property's value
         *
         * @param propertyId The Id for the property to set
         * @param value The new value of the property
         * @sa PropertyValue
         */
        void SetProperty(const uint32_t propertyId, const PropertyValue& value);

        /**
         * @brief Returns a property transaction class, that allows you to set a property and
         * wait for expected side-effects to happen.
         *
         * Setting some properties might result in other properties changing their specification.
         * This method returns a @ref PropertyTransaction object that allows you to set a property
         * and the wait for its expected side-effects on other property specifications.
         *
         * @param propertyId The ID of the property to set, and which triggers side-effects
         * @param value The value to set the property to
         * @return A transaction object that sets the property, and allows waiting for side-effects
         */
        PropertyTransaction BeginSetProperty(const uint32_t propertyId, const PropertyValue& value);

        // Property Discovery

        /**
         * @brief Return a specification object for a given property.
         *
         * @param propertyId The Id of the `Property` to get
         * @return The specification of the provided PropertyId
         */
        PropertySpecification GetPropertySpec(const uint32_t propertyId) const;

        /**
         * @brief Returns a list of all the properties that this camera support.
         *
         * The list of properties are subject to change, and a property
         * may disappear afterwards - not from the returned list, but from the camera.
         * Similarly, a new property may arrive after this function has been called.
         *
         * @return A list of PropertyId's
         * @sa PropertyId
         */
        std::vector<uint32_t> GetAllPropertyIds() const;

        /**
         * @brief Returns the type of tethering used to communicate with this camera
         *
         * This (Windows only) helper method allows you to inspect the established USB link version.
         * (That is; is USB3 used or has the link downgraded to USB2?)
         *
         * For non-Windows platforms, USB2 or USB3 links are indistingushable,
         * and retunred as just 'USB'.
         */
        P1::CameraSdk::TetherConnection GetConnectionType() const;
    };
}
}

#endif  // P1CAMERA_CAMERA_HPP
