#ifndef P1CAMERA_LISTENER_HPP
#define P1CAMERA_LISTENER_HPP

#include <cstdint>
#include <memory>

#include "C_P1CameraCommonProperties.h"
#include "C_P1CameraCommonEvents.h"
#include "C_P1CameraCommonStructs.h"
#include "C_P1CameraCommonLinkage.h"

namespace P1
{
namespace CameraSdk
{
    class Camera;
    class PimplListener;

    /**
     * @brief Class that allows you to receive notifications for events
     *
     * The Listener is notification subscription and receive context. You enable
     * notifications on a Listener object, and wait for incoming notifications on
     * the same Listener.
     *
     * You can create multiple listener objects, and enable different sets of
     * notifications on them. Each listener will receive only notifications
     * explicitly enable on each one of them.
     *
     * One exception is if you enable "catch all" notifications on the listener.
     * Then all notification generated by the CameraSDK, will be sent to the listener,
     * even without a single call to EnableNotification(...).
     *
     * When disable the "catch all" state on the listener, it returns to the normal
     * operating state, delivering only enabled notifications.
     */
    class Listener
    {
    private:
        // MSVC: We cannot dllexport the entire class, since std::shared_ptr<> does 
        // not have a dllexport'ed interface. Therefore, we must explicitly export 
        // every public member manually.
        std::shared_ptr<PimplListener> mImpl;

    public:

        /**
         * Create an empty Listener object, with no enabled notifications
         */
        EXPORT_API Listener();

        /**
         * Destroys a listener, releasing all waiting threads and disabling all notifications
         */
        EXPORT_API ~Listener();

        /**
         * @brief Enables receiving a specific property notification
         *
         * All properties belongs to a @ref Camera. Therefore you must provide the camera from
         * which you would like to receive notifications.
         *
         * Further, there are two kind of property changes that can be enabled: value change
         * and specification changes. Value changes are changes to a property's current value.
         * Specification changes are changes to the @ref PropertySpecification object, and may
         * changes such as read-only, range and value list state.
         *
         * The default options is to enable value change notifications.
         *
         * @param camera The camera who's property we wish to receive notifications for
         * @param propertyId The Id of the property we wish to receive notification for
         * @param isSpecificationChange Set to `true` if you want only specification state change notifications
         */
        EXPORT_API void EnableNotification(const Camera& camera, const uint32_t propertyId, bool isSpecificationChange = false);

        /**
         * @brief Enables receiving notifications for Camera events
         *
         * @param camera The camera to enable the camera event notification on
         * @param eventId The CameraEventId enum for the event type to receive notifications for
         */
        EXPORT_API void EnableNotification(const Camera& camera, const CameraEventId eventId);

        /**
         * @brief Enables receiving notifications for SDK related events
         *
         * @param sdkEventId The SdkEventId enum for the event type to receive notifications for
         */
        EXPORT_API void EnableNotification(const SdkEventId sdkEventId);

        /**
         * @brief Disable receiving property change notifications
         *
         * This causes the listener to stop receiving new notification for the specific
         * property change event.
         *
         * @param camera The camera who's property to stop receiving notifications for
         * @param propertyId The Id of the property to stop receiving notification for
         * @param isSpecificationChange Set to `true` to disable specification change notifications, `false` for value change
         */
        EXPORT_API void DisableNotification(const Camera& camera, const uint32_t propertyId, bool isSpecificationChange = false);

        /**
         * @brief Disable receiving camera event notifications
         */
        EXPORT_API void DisableNotification(const Camera& camera, const CameraEventId eventId);

        /**
         * @brief Disable receiving SDK event notifications
         */
        EXPORT_API void DisableNotification(const SdkEventId sdkEventId);

        /**
         * @brief Enable "catch all" state on this listener
         *
         * In "catch all" state, every notification generated by the SDK is
         * delivered to this listener. (Across all opened cameras!)
         *
         * This also has the effect of enabling notifications for properties that are
         * not currently visible on the cameras. (Due to the cameras state.)
         */
        EXPORT_API void EnableCatchAllNotifications();

        /**
         * @brief Disable "catch all" state on the listener
         *
         * This deactivates the delivery of all types of notifications to this listener.
         * The listener state is returned to normal, where only enabled notification
         * will be delivered. (The ones enabled calling @ref EnableNotification)
         *
         * Any explicitly enabled notifications will be preserved, and are thus not
         * affected by disabling "catch all" state.
         */
        EXPORT_API void DisableCatchAllNotifications();

        /**
         * @brief Block until the next notification is received
         *
         * @param arrivedNotification Out-parameter for the notification that is received, if any
         * @param timeoutMs Time in milliseconds to wait for next notification, before giving up
         * @return `true` if a notification was received, `false` if timed out
         */
        EXPORT_API bool WaitForNotification(NotificationEvent& arrivedNotification, uint32_t timeoutMs = 0);

        /**
         * @brief Wakes up any thread that may be bocked in the WaitForNotification call
         */
        EXPORT_API void WakeUpWaitingThread();
    };
}
}

#endif  // P1CAMERA_LISTENER_HPP
